import logging
import re
import os
import json
from typing import AsyncIterable, Iterable, Optional, Tuple, Dict, List
from urllib.parse import urlparse

from citation_crawler import Crawler, Author, Paper
from .common import download_item, getenv_int

logger = logging.getLogger("semanticscholar")


class SSAuthor(Author):
    def __init__(self, data) -> None:
        super().__init__()
        self.data = data
        assert 'authorId' in self.data and self.data['authorId']

    def authorId(self) -> str:
        return self.data['authorId']

    def name(self) -> Optional[str]:
        if 'name' in self.data:
            return self.data['name']

    def dblp_pid(self) -> Optional[str]:
        return None

    def homepage(self) -> Optional[str]:
        if 'homepage' in self.data:
            return self.data['homepage']

    def dblp_name(self) -> Optional[List[str]]:
        if 'externalIds' in self.data and self.data['externalIds'] and 'DBLP' in self.data['externalIds']:
            return self.data['externalIds']['DBLP']

    def __dict__(self) -> dict:
        d = {}
        if self.authorId():
            d['authorId'] = self.authorId()
        if self.name():
            d['name'] = self.name()
        if self.dblp_name():
            d['dblp_name'] = self.dblp_name()
        return d


async def download_list(url: str, path: str, cache_days: int):
    def list_data_is_valid(text):
        data = json.loads(text)
        if 'data' in data:
            return True
        raise ValueError(f"Invalid list data: {text}")
    text = await download_item(url, path, cache_days, list_data_is_valid)
    if not text:
        return
    return json.loads(text)


fields_authors = "externalIds,name,affiliations,homepage"
root_authors = f"semanticscholar/authors--{fields_authors.replace(',', '-')}"


async def get_authors(paperId: str) -> Iterable[Author]:
    cache_days = getenv_int('CITATION_CRAWLER_MAX_CACHE_DAYS_AUTHORS')
    cache_days = cache_days if cache_days is not None else -1
    paperId = paperId.lower()
    url = f"https://api.semanticscholar.org/graph/v1/paper/{paperId}/authors?fields={fields_authors}"
    data = await download_list(url, os.path.join(root_authors, f"{paperId}.json"), cache_days)
    if not data or 'data' not in data:
        return
    for a in data['data']:
        if 'authorId' in a and a['authorId']:
            yield SSAuthor(a)


class SSPaper(Paper):
    def __init__(self, data) -> None:
        super().__init__()
        self.data = data
        self.author_data = None
        assert 'paperId' in self.data and self.data['paperId']

    def paperId(self) -> str:
        return self.data['paperId']

    def dblp_id(self) -> Optional[str]:
        if 'externalIds' in self.data and 'DBLP' in self.data['externalIds']:
            return self.data['externalIds']['DBLP']

    def title(self) -> Optional[str]:
        if 'title' in self.data:
            return self.data['title']

    def year(self) -> Optional[int]:
        if 'year' in self.data:
            return self.data['year']

    def date(self) -> Optional[int]:
        if 'publicationDate' in self.data:
            return self.data['publicationDate']

    def doi(self) -> Optional[str]:
        if 'externalIds' in self.data and self.data['externalIds'] and 'DOI' in self.data['externalIds']:
            doi = self.data['externalIds']['DOI']
            u = urlparse(doi)
            doi = re.sub(r"^/+", "", u.path)
            return doi

    def abstract(self) -> Optional[int]:
        if 'abstract' in self.data:
            return self.data['abstract']

    async def _get_authors_from_author_data(self) -> Iterable[Author]:
        if not self.author_data:
            authors = []
            async for author in get_authors(self.paperId()):
                authors.append(author)
            self.author_data = authors
        for author in self.author_data:
            yield author

    async def authors(self) -> Iterable[SSAuthor]:
        if 'authors' in self.data and len(self.data['authors']) >= 0:
            for a in self.data['authors']:
                if 'authorId' not in a or 'externalIds' not in a or not a['externalIds']:
                    async for author in self._get_authors_from_author_data():
                        yield author
                    return
            for a in self.data['authors']:
                yield SSAuthor(a)
        else:
            async for author in self._get_authors_from_author_data():
                yield author

    async def authors_kv(self) -> Iterable[Tuple[str, str]]:
        async for author in self.authors():
            if author.dblp_pid():
                yield "dblp_pid", author.dblp_pid()
            if author.authorId():
                yield "authorId", author.authorId()

    async def get_references(self) -> Iterable[Paper]:
        async for paper in get_references(self.paperId()):
            yield paper

    async def get_citations(self) -> Iterable[Paper]:
        async for paper in get_citations(self.paperId()):
            yield paper


fields_references = f"title,abstract,year,authors,externalIds,publicationTypes,journal"
root_references = f"semanticscholar/references--{fields_references.replace(',', '-')}"
root_citations = f"semanticscholar/citations--{fields_references.replace(',', '-')}"


async def get_references(paperId: str) -> Iterable[SSPaper]:
    cache_days = getenv_int('CITATION_CRAWLER_MAX_CACHE_DAYS_REFERENCES')
    cache_days = cache_days if cache_days is not None else -1
    paperId = paperId.lower()
    url = f"https://api.semanticscholar.org/graph/v1/paper/{paperId}/references?fields={fields_references}"
    data = await download_list(url, os.path.join(root_references, f"{paperId}.json"), cache_days)
    if not data or 'data' not in data:
        return
    data = data['data']
    for d in data:
        if 'citedPaper' in d and 'paperId' in d['citedPaper'] and d['citedPaper']['paperId']:
            yield SSPaper(d['citedPaper'])


async def get_citations(paperId: str) -> Iterable[SSPaper]:
    cache_days = getenv_int('CITATION_CRAWLER_MAX_CACHE_DAYS_CITATIONS')
    cache_days = cache_days if cache_days is not None else 7
    paperId = paperId.lower()
    url = f"https://api.semanticscholar.org/graph/v1/paper/{paperId}/citations?fields={fields_references}"
    data = await download_list(url, os.path.join(root_citations, f"{paperId}.json"), cache_days)
    if not data or 'data' not in data:
        return
    data = data['data']
    for d in data:
        if 'citingPaper' in d and 'paperId' in d['citingPaper'] and d['citingPaper']['paperId']:
            yield SSPaper(d['citingPaper'])


async def download_paper(url: str, path: str, cache_days: int):
    def paper_is_valid(text):
        data = json.loads(text)
        if 'paperId' in data:
            return True
        raise ValueError(f"Invalid paper data: {text}")
    text = await download_item(url, path, cache_days, paper_is_valid)
    if not text:
        return
    return json.loads(text)


fields_authors_sub = ','.join([("authors." + f) for f in fields_authors.split(',')])
fields_paper = f"title,abstract,year,publicationDate,{fields_authors_sub},externalIds,publicationTypes,journal"
root_paper = f"semanticscholar/paper--{fields_paper.replace(',', '-')}"


async def get_paper(paperId: str) -> Optional[SSPaper]:
    def paperId2path(paperId):
        return paperId.replace(":", "/")
    cache_days = getenv_int('CITATION_CRAWLER_MAX_CACHE_DAYS_PAPER')
    cache_days = cache_days if cache_days is not None else -1
    paperId = paperId.lower()
    url = f"https://api.semanticscholar.org/graph/v1/paper/{paperId}?fields={fields_paper}"
    data = await download_paper(url, os.path.join(root_paper, f"{paperId2path(paperId)}.json"), cache_days)
    if not data or 'paperId' not in data or data['paperId'].lower() != paperId:
        return None
    return SSPaper(data)


fields_author = "paperId,title"
root_author = f"semanticscholar/author--{fields_author.replace(',', '-')}"


async def get_paperIds_by_authorId(authorId: str) -> List[str]:
    cache_days = getenv_int('CITATION_CRAWLER_MAX_CACHE_DAYS_INIT_AUTHOR')
    cache_days = cache_days if cache_days is not None else 7
    authorId = authorId.lower()
    url = f"https://api.semanticscholar.org/graph/v1/author/{authorId}/papers?fields={fields_author}&limit=100"
    data = await download_list(url, os.path.join(root_author, f"{authorId}.json"), cache_days)
    if not data or 'data' not in data:
        return
    for paper in data['data']:
        yield paper['paperId'].lower()


class SemanticScholarCrawler(Crawler):

    def __init__(self, authorId_list: List[str], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.authors = authorId_list

    async def get_init_paperIds(self):
        for author in self.authors:
            async for paperId in get_paperIds_by_authorId(author):
                yield paperId

    async def get_paper(self, paperId):
        return await get_paper(paperId)

    async def get_references(self, paper):
        async for paper in get_references(paper.paperId()):
            yield paper

    async def get_citations(self, paper):
        async for paper in get_citations(paper.paperId()):
            yield paper

    async def match_authors(self, paper: SSPaper, authors: AsyncIterable[Dict]) -> AsyncIterable[Tuple[Dict, Dict, bool]]:
        dblp_names, authorIds = {}, {}
        async for author in paper.authors():
            if author.dblp_name():
                for name in author.dblp_name():
                    dblp_names[name] = author
            if author.authorId():
                authorIds[author.authorId()] = author
        async for author in authors:
            if 'authorId' in author:
                if author['authorId'] in authorIds:
                    ss_author = authorIds[author['authorId']]
                    write_fields = {}
                    if ss_author.homepage():
                        write_fields["homepage"] = ss_author.homepage()
                    author_kv = {"authorId": author['authorId']}
                    yield author_kv, write_fields, None
                else:  # if there is an author in database but is not really an author
                    # should unlink the author
                    division_kv = {"authorId": author['authorId']}
                    if 'name' in author and author['name'] in dblp_names:
                        ss_author = dblp_names[author['name']]
                        author_kv = {"authorId": ss_author.authorId()}
                        yield author_kv, {}, division_kv
            elif 'name' in author and author['name'] in dblp_names:
                ss_author = dblp_names[author['name']]
                write_fields = {"authorId": ss_author.authorId()}
                if ss_author.homepage():
                    write_fields["homepage"] = ss_author.homepage()
                author_kv = {"dblp_pid": author["dblp_pid"]}
                yield author_kv, write_fields, None



import abc
import logging
import asyncio
from tqdm.asyncio import tqdm
from typing import Tuple, Optional, AsyncIterable, List
import random
from dblp_crawler.gather import gather
from .items import Paper


logger = logging.getLogger("graph")


class Summarizer(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    async def filter_papers(self, papers: AsyncIterable[Paper]) -> AsyncIterable[Paper]:
        """
        在输出时过滤`Paper`，被过滤掉的`Paper`将不会出现在输出中
        等同于dblp-crawler里的filter_publications_at_output
        """
        async for paper in papers:
            yield paper

    @abc.abstractmethod
    async def write_paper(self, paper: Paper) -> None:
        pass

    @abc.abstractmethod
    async def write_reference(self, paper: Paper, reference: Paper) -> None:
        pass

    @abc.abstractmethod
    async def get_corrlated_authors(self, paper: Paper) -> AsyncIterable[dict]:
        """
        Return existing authors (in dict format) of a paper in database
        用于将数据库中的作者与爬到的作者进行匹配
        """
        async for author in paper.authors():
            yield author.__dict__()

    @abc.abstractmethod
    async def write_author(self, paper: Paper, author_kv: dict, write_fields: dict, division_kv: bool) -> None:
        pass


class Crawler(metaclass=abc.ABCMeta):
    def __init__(self, summarizer: Summarizer, paperId_list: List[str]) -> None:
        self.summarizer = summarizer
        self._init_paper_list = paperId_list
        self.papers: Dict[str, Paper] = {}
        self.fetched = set()
        self.ref_idx: Dict[str, set[str]] = {}
        self.inited = False

    @abc.abstractmethod
    async def get_init_paperIds(self) -> AsyncIterable[str]:
        """初始化"""
        return

    @abc.abstractmethod
    async def get_paper(self, paperId: str) -> Optional[Paper]:
        """获取某篇论文的详情"""
        return None

    @abc.abstractmethod
    async def get_references(self, paper: Paper) -> AsyncIterable[Paper]:
        """获取某篇论文的引文"""
        return

    @abc.abstractmethod
    async def get_citations(self, paper: Paper) -> AsyncIterable[Paper]:
        """获取引用某篇论文的论文"""
        return

    @abc.abstractmethod
    async def filter_papers(self, papers: AsyncIterable[Paper]) -> AsyncIterable[Paper]:
        """
        在收集信息时过滤`Paper`，不会对被此方法过滤掉的`Paper`进行信息收集
        等同于dblp-crawler里的filter_publications_at_crawler
        """
        async for paper in papers:
            yield paper

    @abc.abstractmethod
    async def match_authors(self, paper: Paper, authors: AsyncIterable[dict]) -> AsyncIterable[Tuple[dict, dict, bool]]:
        """
        This function will be called after Summarizer.get_corrlated_authors
        Return matched authors and the fields you want to write (in dict format)
        用于将数据库中的作者与爬到的作者进行匹配
        """
        async for author in authors:
            yield author, author

    async def init_paper(self, paperId) -> Tuple[Optional[Paper], int]:
        # fetch论文
        if paperId not in self.papers:  # init时self.papers里肯定没有数据
            paper = await self.get_paper(paperId)
            if not isinstance(paper, Paper):
                return None, 0
            paperId = paper.paperId()
        else:  # init之后的文章肯定作为references或citations已经下载过了
            paper = self.papers[paperId]
            paperId = paper.paperId()
        self.papers[paperId] = paper

        # fetch references
        refs, cits = 0, 0
        async for new_paper in self.filter_papers(self.get_references(paper)):
            if not new_paper:
                continue
            new_paperId = new_paper.paperId()
            if paperId not in self.ref_idx:
                self.ref_idx[paperId] = set()
            self.ref_idx[paperId].add(new_paperId)
            if new_paperId not in self.papers or not self.papers[new_paperId]:
                self.papers[new_paperId] = new_paper
                refs += 1

        # fetch citations
        async for new_paper in self.filter_papers(self.get_citations(paper)):
            if not new_paper:
                continue
            new_paperId = new_paper.paperId()
            if new_paperId not in self.ref_idx:
                self.ref_idx[new_paperId] = set()
            self.ref_idx[new_paperId].add(paperId)
            if new_paperId not in self.papers or not self.papers[new_paperId]:
                self.papers[new_paperId] = new_paper
                cits += 1

        logger.info("There are %s refernces and %s citations in %s" % (refs, cits, paperId))
        return paper, refs + cits

    async def _init_papers(self):
        tasks = []
        for paperId in self._init_paper_list:
            if paperId in self.fetched:
                continue
            self.fetched.add(paperId)
            logger.info("Init paper: %s" % paperId)
            tasks.append(self.init_paper(paperId))
        async for paperId in self.get_init_paperIds():
            if paperId in self.fetched:
                continue
            self.fetched.add(paperId)
            logger.info("Init paper: %s" % paperId)
            tasks.append(self.init_paper(paperId))
        random.shuffle(tasks)
        async for paper, news in tqdm(gather(*tasks), desc="Writing init papers", total=len(tasks)):
            yield paper, news

    async def _bfs_once(self) -> int:
        # 初始化
        if not self.inited:
            async for paper, news in self._init_papers():
                if isinstance(paper, Paper):
                    yield paper, news
            self.inited = True

        # 构造待fetch论文列表
        paperIds = []
        for paperId in list(self.papers.keys()):
            if paperId in self.fetched:
                continue
            self.fetched.add(paperId)
            paperIds.append(paperId)
            logger.info("Fetch paper: %s" % paperId)

        # 执行fetch论文
        tasks = [self.init_paper(paperId) for paperId in paperIds]
        random.shuffle(tasks)
        async for paper, news in tqdm(gather(*tasks), desc="Writing papers", total=len(tasks)):
            if isinstance(paper, Paper):
                yield paper, news

    async def bfs_once(self) -> None:
        total, total_news = 0, 0
        async for paper, news in self.summarizer.filter_papers(self._bfs_once()):
            total += 1
            total_news += news
            await self.summarizer.write_paper(paper)  # _bfs_once里面出来的每个paper都是新的，所以直接写入
            async for author_kv, write_fields, division_kv in self.match_authors(paper, self.summarizer.get_corrlated_authors(paper)):
                # _bfs_once里面出来的每个paper都是新的，所以直接写入
                await self.summarizer.write_author(paper, author_kv, write_fields, division_kv)
            for paperId, refs_paperId in list(self.ref_idx.items()):
                # _bfs_once里面出来的paper不能保证引文全部已获取到
                for ref_paperId in list(refs_paperId):
                    if not (paperId == paper.paperId() or ref_paperId == paper.paperId()):
                        continue  # 只写入相关的论文引文
                    if not (paperId in self.papers and ref_paperId in self.papers):
                        continue  # 只写入已入库的论文引文
                    await self.summarizer.write_reference(self.papers[paperId], self.papers[ref_paperId])
                    refs_paperId.remove(ref_paperId)  # 删除已入库的论文引文
                    if len(refs_paperId) <= 0:
                        del self.ref_idx[paperId]  # 删除已入库的论文引文
        logger.info("Fetched %d papers from %d papers" % (total_news, total))
        return total_news



from typing import Optional, Dict, Callable
import os
from datetime import datetime, timedelta

import aiohttp
import logging
from aiofile import async_open
from asyncio import Semaphore

logger = logging.getLogger("common")


def getenv_int(key) -> int:
    cache_days = os.getenv(key)
    if cache_days is not None:
        try:
            return int(cache_days)
        except:
            pass
    return None


http_concorent = getenv_int('HTTP_CONCORRENT')
http_sem = Semaphore(http_concorent if http_concorent is not None else 8)
file_sem = Semaphore(512)


def get_cache_datetime(path) -> datetime:
    return datetime.fromtimestamp(os.path.getmtime(path))


async def download_item(url: str, path: str, cache_days: int, is_valid: Callable[[str], None]) -> Optional[Dict]:
    save_path = os.path.join("save", path)
    if os.path.isfile(save_path):
        if cache_days < 0 or datetime.now() < get_cache_datetime(save_path) + timedelta(days=cache_days):
            async with file_sem:
                try:
                    async with async_open(save_path, 'r') as f:
                        logger.debug("use cache: %s -> %s" % (path, url))
                        text = await f.read()
                    assert is_valid(text)
                    return text
                except:
                    logger.debug(" no cache: %s" % save_path)
                    if os.path.exists(save_path):
                        logger.info("err cache: %s" % save_path)
                        try:
                            os.remove(save_path)
                        except Exception as e:
                            logger.info(" rm cache: %s" % e)
        else:
            logger.info("old cache: %s" % save_path)

    async with http_sem:
        try:
            async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(verify_ssl=False)) as session:
                async with session.get(url,
                                       proxy=os.getenv("HTTP_PROXY"),
                                       timeout=os.getenv("HTTP_TIMEOUT") or 30) as response:
                    logger.debug(" download: %s <- %s" % (path, url))
                    text = await response.text()
                    assert is_valid(text)
                    os.makedirs(os.path.dirname(save_path), exist_ok=True)
                    async with async_open(save_path, 'w') as f:
                        await f.write(text)
                    return text
        except Exception as e:
            logger.error(" down err: %s" % e)
    return None



import logging
from typing import AsyncIterable
from citation_crawler import Summarizer, Paper

import dateutil.parser
from neo4j import AsyncSession
import neo4j.time

'''Use with dblp-crawler'''

logger = logging.getLogger("graph")


async def add_paper(tx, paper: Paper):
    n4jset = "MERGE (p:Publication {title_hash: $title_hash}) "\
        "SET p.title=$title, p.year=$year"
    if paper.doi():
        n4jset += ", p.doi=$doi"
    if paper.abstract():
        n4jset += ", p.abstract=$abstract"
    if paper.dblp_id():
        n4jset += ", p.dblp_key=$dblp_id"
    if paper.paperId():
        n4jset += ", p.paperId=$paperId"
    date = None
    if paper.date():
        try:
            _date = dateutil.parser.parse(paper.date())
            date = neo4j.time.Date(
                year=_date.year,
                month=_date.month,
                day=_date.day,
            )
            n4jset += ", p.date=$date"
        except Exception as e:
            logger.error(f"Cannot parse date {paper.date()}: {e}")
    await tx.run(n4jset,
                 title_hash=paper.title_hash(),
                 title=paper.title(),
                 year=paper.year(),
                 paperId=paper.paperId(),
                 dblp_id=paper.dblp_id(),
                 doi=paper.doi(),
                 abstract=paper.abstract(),
                 date=date)


async def add_reference(tx, a: Paper, b: Paper):
    await tx.run("MERGE (a:Publication {title_hash: $a}) "
                 "MERGE (b:Publication {title_hash: $b}) "
                 "MERGE (a)-[:CITE]->(b)",
                 a=a.title_hash(), b=b.title_hash())


async def match_corrlated_authors(tx, paper: Paper):
    nodes = []
    for record in await (await tx.run("MATCH (a:Person)-[:WRITE]->(p:Publication {title_hash: $title_hash}) return a",
                                      title_hash=paper.title_hash())).values():
        nodes.append({**dict(record[0]), "element_id": record[0].element_id})
    return nodes


async def match_authors_kv(tx, k, v):
    nodes = []
    for record in await (await tx.run("MATCH (a:Person {%s: $value}) RETURN a" % k,
                                      value=v)).values():
        nodes.append({**dict(record[0]), "element_id": record[0].element_id})
    return nodes


async def link_author(tx, paper: Paper, author_kv, write_fields):
    await tx.run("MERGE (p:Publication {title_hash: $title_hash}) " +
                 ('MERGE (a:Person {%s}) ' % (",".join([f'{k}: ${k}' for k in author_kv]))) +
                 (f'SET {",".join([f"a.{k}=${k}" for k in write_fields])}' if len(write_fields) > 0 else "") +
                 " MERGE (a)-[:WRITE]->(p)",
                 title_hash=paper.title_hash(), **author_kv, **write_fields)


async def divide_author(tx, paper: Paper, author_kv, write_fields, division_kv):
    author_c = (await (await tx.run("MATCH (c:Person {%s}) RETURN c" % (",".join([f'{k}: ${k}' for k in division_kv])),
                                    **division_kv)).values())[0][0]._properties
    await tx.run("MATCH (c:Person {%s})-[r:WRITE]->(p:Publication {title_hash: $title_hash}) "
                 "DELETE r" % (",".join([f'{k}: ${k}' for k in division_kv])),
                 title_hash=paper.title_hash(), **division_kv)
    write_fields = {**author_c, **write_fields, **author_kv}
    await tx.run("MERGE (p:Publication {title_hash: $title_hash}) " +
                 ('MERGE (a:Person {%s}) ' % (",".join([f'{k}: ${k}' for k in author_kv]))) +
                 (f'SET {",".join([f"a.{k}=${k}" for k in write_fields])}' if len(write_fields) > 0 else "") +
                 " MERGE (a)-[:WRITE]->(p)",
                 title_hash=paper.title_hash(), **write_fields)


async def _add_references(tx, paper: Paper, skip_exists=True):
    title_hash_exists = set([
        title_hash for (title_hash,) in
        await (await tx.run("MATCH (a:Publication)-[:CITE]->(p:Publication {title_hash: $title_hash}) RETURN a.title_hash",
               title_hash=paper.title_hash())).values()
    ])
    async for ref in paper.get_references():
        if skip_exists and ref.title_hash() in title_hash_exists:
            continue
        await add_paper(tx, ref)
        await add_reference(tx, paper, ref)


async def _add_citations(tx, paper: Paper, skip_exists=True):
    title_hash_exists = set([
        title_hash for (title_hash,) in
        await (await tx.run("MATCH (p:Publication {title_hash: $title_hash})-[:CITE]->(a:Publication) RETURN a.title_hash",
               title_hash=paper.title_hash())).values()
    ])
    async for cit in paper.get_references():
        if skip_exists and cit.title_hash() in title_hash_exists:
            continue
        await add_paper(tx, cit)
        await add_reference(tx, cit, paper)


class Neo4jSummarizer(Summarizer):
    def __init__(self, session: AsyncSession, skip_exists=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = session
        self.skip_exists = skip_exists

    async def write_paper(self, paper) -> None:
        await self.session.execute_write(add_paper, paper)
        await self.session.execute_write(_add_references, paper, self.skip_exists)
        await self.session.execute_write(_add_citations, paper, self.skip_exists)

    async def write_reference(self, paper, reference) -> None:
        await self.session.execute_write(add_reference, paper, reference)

    async def get_corrlated_authors(self, paper: Paper) -> AsyncIterable[dict]:
        authors = set()
        for author in await self.session.execute_read(match_corrlated_authors, paper):
            if author["element_id"] not in authors:
                yield author
                authors.add(author["element_id"])
        async for k, v in paper.authors_kv():
            for author in await self.session.execute_read(match_authors_kv, k, v):
                if author["element_id"] not in authors:
                    yield author
                    authors.add(author["element_id"])

    async def write_author(self, paper: Paper, author_kv, write_fields, division_kv) -> None:
        if division_kv:
            await self.session.execute_write(divide_author, paper, author_kv, write_fields, division_kv)
        await self.session.execute_write(link_author, paper, author_kv, write_fields)



import logging
import json
from citation_crawler import Summarizer


import networkx as nx

'''Use with dblp-crawler'''

logger = logging.getLogger("graph")


class NetworkxSummarizer(Summarizer):
    def __init__(self: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.graph: nx.DiGraph = nx.DiGraph()

    async def write_paper(self, paper) -> None:
        self.graph.add_node(paper.paperId(), paper=paper)

    async def write_author(self, paper, author_dict, write_fields, division):
        for _ in []:
            yield None  # yet do not write author in networkx

    async def get_corrlated_authors(self, paper):
        for _ in []:
            yield None  # yet do not write author in networkx

    async def write_reference(self, paper, reference) -> None:
        self.graph.add_node(paper.paperId(), paper=paper)
        self.graph.add_node(reference.paperId(), paper=reference)
        self.graph.add_edge(paper.paperId(), reference.paperId())

    async def save(self, jsonpath) -> None:
        nodes = {}
        for k, d in self.graph.nodes(data=True):
            if k not in nodes:
                nodes[k] = await d["paper"].__dict__()
            else:
                nodes[k] = {**nodes[k], **await d["paper"].__dict__()}
        edges = [(u, v) for u, v in self.graph.edges()]
        with open(jsonpath, 'w', encoding="utf8") as f:
            json.dump(dict(nodes=nodes, edges=edges), f, indent=2)



from neo4j import GraphDatabase
from urllib.parse import urlparse
import re


def match_papers(tx):
    papers = []
    for record in tx.run("MATCH (p:Publication) RETURN p.paperId, p.doi"):
        paperId, doi = record.values()
        if paperId is not None:
            papers.append(paperId)
        elif doi is not None:
            u = urlparse(doi)
            papers.append("DOI:" + re.sub(r"^/+", "", u.path))
    return papers


def papers_in_neo4j(uri, auth=None):
    with GraphDatabase.driver(uri, auth=auth) as driver:
        with driver.session() as session:
            return session.execute_read(match_papers)


def match_authors(tx):
    authors = []
    for record in tx.run("MATCH (p:Person) RETURN p.authorId"):
        for value in record.values():
            if value is not None:
                authors.append(value)
    return authors


def authors_in_neo4j(uri, auth=None):
    with GraphDatabase.driver(uri, auth=auth) as driver:
        with driver.session() as session:
            return session.execute_read(match_authors)


def match_papers_keywords(tx, year, *arg_keywords):
    pwhere = 'WHERE p.year >= $year'
    values = dict(year=year)

    ki, k_or, v_or = 0, [], {}
    for keywords in arg_keywords:
        k_and, v_and = [], {}
        for k in keywords.split(" "):
            if not k:
                continue
            ki += 1
            k_and.append(f"toLower(p.title) CONTAINS $keyword{ki}")
            v_and[f"keyword{ki}"] = k
        k_or.append(f"({' and '.join(k_and)})")
        v_or = {**v_or, **v_and}
    if ki > 0:
        pwhere += " AND " + f"({' OR '.join(k_or)})"
        values = {**values, **v_or}

    papers = []
    for record in tx.run("MATCH (p:Publication)" + pwhere + " RETURN p.paperId, p.doi", **values):
        paperId, doi = record.values()
        if paperId is not None:
            papers.append(paperId)
        elif doi is not None:
            u = urlparse(doi)
            papers.append("DOI:" + re.sub(r"^/+", "", u.path))
    return papers


def papers_in_neo4j_keywords(uri, auth=None, year=2000, *keywords):
    with GraphDatabase.driver(uri, auth=auth) as driver:
        with driver.session() as session:
            return session.execute_read(match_papers_keywords, year, *keywords)



import abc
import re
from typing import Optional, Iterable, Tuple
from typing_extensions import Self


class Author(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def authorId(self) -> str:
        return ''

    @abc.abstractmethod
    def name(self) -> Optional[str]:
        return None

    @abc.abstractmethod
    def dblp_pid(self) -> Optional[str]:
        return None

    def __dict__(self) -> dict:
        d = {}
        if self.authorId():
            d['authorId'] = self.authorId()
        if self.name():
            d['name'] = self.name()
        if self.dblp_pid():
            d['dblp_pid'] = self.dblp_pid()
        return d


class Paper(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def paperId(self) -> str:
        return ''

    @abc.abstractmethod
    def dblp_id(self) -> Optional[str]:
        return None

    @abc.abstractmethod
    def title(self) -> str:
        return None

    def title_hash(self) -> str:
        return re.sub(r"[^0-9a-z]", "", self.title().lower())

    @abc.abstractmethod
    def year(self) -> Optional[int]:
        return None

    @abc.abstractmethod
    def date(self) -> Optional[str]:
        return None

    @abc.abstractmethod
    def doi(self) -> Optional[str]:
        return None

    @abc.abstractmethod
    def abstract(self) -> Optional[str]:
        return None

    @abc.abstractmethod
    async def authors(self) -> Iterable[Author]:
        return

    @abc.abstractmethod
    async def authors_kv(self) -> Iterable[Tuple[str, str]]:
        """key and correlated value to match authors"""
        return

    @abc.abstractmethod
    async def get_references(self) -> Iterable[Self]:
        return

    @abc.abstractmethod
    async def get_citations(self) -> Iterable[Self]:
        return

    async def __dict__(self) -> dict:
        d = {}
        if self.paperId():
            d['paperId'] = self.paperId()
        if self.dblp_id():
            d['dblp_key'] = self.dblp_id()
        if self.title():
            d['title'] = self.title()
        if self.title_hash():
            d['title_hash'] = self.title_hash()
        if self.year():
            d['year'] = self.year()
        if self.date():
            d['date'] = self.date()
        if self.doi():
            d['doi'] = self.doi()
        if self.abstract():
            d['abstract'] = self.abstract()
        d['authors'] = []
        async for author in self.authors():
            d['authors'].append(author.__dict__())
        return d
